﻿using FLY.Modbus;
using FLY.OBJComponents.Common;
using FLY.OBJComponents.IService;
using FLY.OBJComponents.Server;
using FLY.Winder.Common;
using FLY.Winder.IService;
using FLY.Winder.Server.Model;
using FObjBase;
using Misc;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;

namespace FLY.Winder.Server
{
    public class WinderSystem : IWinderSystemService
    {
        #region 延时推送 MARKNO
        const int MARKNO_SAVE = 1;
        const int MARKNO_DELAY_ISCONNECTED = 4;
        #endregion
        /// <summary>
        /// 收卷其它附件
        /// </summary>
        public WinderAccessory Accessory { get; private set; }

        /// <summary>
        /// 内收卷 外收卷
        /// </summary>
        public List<WinderInsideOutside> Items { get; private set; }

        private PLCProxySystem plcos = new PLCProxySystem();
        /// <summary>
        /// PLC代理系统
        /// </summary>
        public IPLCProxySystemService PLCos { get { return plcos; } }

        /// <summary>
        /// 报警系统
        /// </summary>
        WarningSystem mWarning;
        HistoryDb mHistoryDb;


        public WinderSystem()
        {
            //加载PLC寄存器 文件
            AddConfigFile("WS.xml");




        }

        public void Init(HistoryDb historyDb, WarningSystem warning)
        {
            mHistoryDb = historyDb;
            mWarning = warning;

            //--------------------------------------------------------------------------------
            //报警配置
            InitError();

            FObjBase.PollModule.Current.Poll_Config(
                FObjBase.PollModule.POLL_CONFIG.ADD, OnPoll, TimeSpan.FromSeconds(1));

            //注册通知事件
            plcos.Init();
        }
        #region 报警
        class ErrorAction
        {
            public List<ErrorInfo> error_property;
            public object state;
            public delegate void ErrorHandler(ref string msg, object state);
            public ErrorHandler action;
        }
        class ErrorInfo 
        {
            public string property;
            public byte code;
            public string msg;
            public bool offIsError;
        }
        Dictionary<INotifyPropertyChanged, ErrorAction> obj_error = new Dictionary<INotifyPropertyChanged, ErrorAction>();

        void AddErrorAction(INotifyPropertyChanged sender, ref byte code, ErrorAction.ErrorHandler errorHandler=null)
        {
            //反射找出全部是报警的property
            List<ErrorInfo> error_property = new List<ErrorInfo>();
            foreach (var propertyInfo in sender.GetType().GetProperties())
            {
                if (propertyInfo.GetCustomAttributes(typeof(IsErrorAttribute), false).Count() > 0)
                {
                    //肯定有，没有就让它出错！！！
                    string desp = (propertyInfo.GetCustomAttributes(typeof(DescriptionAttribute), false).First() as DescriptionAttribute).Description;
                    bool offIsError = (propertyInfo.GetCustomAttributes(typeof(IsErrorAttribute), false).First() as IsErrorAttribute).OffIsError;


                    ErrorInfo errorInfo = new ErrorInfo()
                    {
                        property = propertyInfo.Name,
                        msg = desp,
                        offIsError = offIsError,
                        code = code
                    };
                    error_property.Add(errorInfo);
                    code++;
                }
            }
            obj_error.Add(sender, new ErrorAction()
            {
                error_property = error_property,
                action = errorHandler
            });
        }
        void InitError()
        {
            byte code = 0;
            //反射找出全部是报警的property
            AddErrorAction(
                Accessory, 
                ref code);

            AddErrorAction(
                Items[0], 
                ref code, 
                (ref string description, object state) =>{ description = "内" + description; }
                );

            AddErrorAction(
                Items[1],
                ref code,
                (ref string description, object state) => { description = "外" + description; }
                );

            foreach (var obj in obj_error.Keys)
            {
                obj.PropertyChanged += Obj_PropertyChanged_ForError;
            }

            //--------------------------------------------------------------------------------
            //添加任务
            foreach (var kv in obj_error)
            {
                string objname = PLCos.ObjNames.First(_kv => _kv.Value == kv.Key).Key;
                PLCos.SetPlan(objname, kv.Value.error_property.Select(ei=>ei.property).ToArray(), 0);
            }

            //--------------------------------------------------------------------------------
            //连接断开事件
            FObjBase.PollModule.Current.Poll_Config(PollModule.POLL_CONFIG.ADD,
                () =>
                {
                    Misc.BindingOperations.SetBinding(plcos, "IsConnectedWithPLC", () =>
                    {
                        bool b = !PLCos.IsConnectedWithPLC;

                        ERR_STATE state = b ? ERR_STATE.ON : ERR_STATE.OFF;

                        ERRNO errno = ERRNOs.ERRNO_PLC_DISCONNECTED;
                        byte errcode = errno.Code;
                        string description = errno.Descrption;
                        mWarning.Add(errcode, description, state);
                    });
                }, TimeSpan.FromSeconds(3), true, false, this, MARKNO_DELAY_ISCONNECTED, true);

            //--------------------------------------------------------------------------------
            //把全部报警状态取反，也就是不报警！！！
            foreach (var kv in obj_error)
            {
                object sender = kv.Key;
                foreach (var ei in kv.Value.error_property) 
                {
                    if (ei.offIsError)
                    {
                        Misc.PropertiesManager.SetValue(sender, ei.property, true);
                    }
                    //false 不用写，默认值就是
                }
            }

        }


        void Obj_PropertyChanged_ForError(object sender, PropertyChangedEventArgs e)
        {
            ErrorAction errorAction = obj_error[sender as INotifyPropertyChanged];

            var  eis = from ei in errorAction.error_property where ei.property == e.PropertyName select ei;
            if(eis.Count()>0)
            {
                var ei = eis.First();

                //获取描述
                var type = sender.GetType();
                var property = type.GetProperty(ei.property);
                bool b = (bool)property.GetValue(sender,null);
                string desp = ei.msg;
                bool offIsError = ei.offIsError;

                ERR_STATE state;
                if(offIsError)
                    state = !b ? ERR_STATE.ON : ERR_STATE.OFF; 
                else
                    state = b ? ERR_STATE.ON : ERR_STATE.OFF;

                byte errcode = ei.code;//每个报警必须不一样！！
                string description = desp;

                errorAction.action?.Invoke(ref description, errorAction.state);

                mWarning.Add(errcode, description, state);
            }
        }
        #endregion


        void AddConfigFile(string configfile)
        {
            PLCGroup plcgroup = new PLCGroup();
            plcgroup.Load(configfile);

            foreach (PLCGroup.PLCDevice device in plcgroup.Devices)
            {
                var plc = new Modbus.WithThread.ModbusMapper_Client(device.EP);
                plcos.PLCs.Add(plc);
            }

            Items = new List<WinderInsideOutside>();
            for (int i = 0; i < 2; i++)
            {
                Items.Add(new WinderInsideOutside());
            }
            Accessory = new WinderAccessory();

            //objname 转 obj
            //PLCos.ObjNames[n].Key 必须与 WS.xml 中的 OwnerName 一致！！！！
            PLCos.ObjNames.Add("Accessory", Accessory);
            for (int i = 0; i < Items.Count(); i++)
                PLCos.ObjNames.Add($"Items[{i}]", Items[i]);


            foreach (PLCGroup.PLCVariable var in plcgroup.Variables)
            {
                if (var.DeviceIndex < 0 || var.DeviceIndex >= plcos.PLCs.Count)
                    continue;

                List<Modbus.DataToRegs> drs = plcos.DRMap;
                var plc = plcos.PLCs[var.DeviceIndex];


                Modbus.DataToRegs dr = plc.MapDataToRegs(
                    ModbusMapper.TranslateToPLCAddressArea(var.Mode),
                    var.Addr,
                    ModbusMapper.TranslateToREG_TYPE(var.Type),
                    var.Scale,
                    PLCos.ObjNames[var.OwnerName],
                    var.PropertyName);
                if (dr != null)
                    drs.Add(dr);
            }

            foreach (var plc in plcos.PLCs)
                plc.Build();
        }


        Stopwatch stopwatch_infoInterval = new Stopwatch();
        Stopwatch stopwatch_1min = new Stopwatch();
        List<Db_WinderInfo> dbInfo_1min = new List<Db_WinderInfo>();
        void OnPoll()
        {
            if (stopwatch_1min.IsRunning && stopwatch_1min.Elapsed >= TimeSpan.FromSeconds(60))
            {
                //启动了1min 且到点了
                if (dbInfo_1min.Count > 0)
                {
                    mHistoryDb.AddWinderInfos(dbInfo_1min);
                    dbInfo_1min.Clear();
                    stopwatch_1min.Restart();
                }
                else 
                {
                    stopwatch_1min.Reset();
                }
            }


            if (!stopwatch_infoInterval.IsRunning)
            {
                stopwatch_infoInterval.Start();
                return;
            }
            else
            {
                if (stopwatch_infoInterval.Elapsed < TimeSpan.FromSeconds(10))
                    return;
                else
                    stopwatch_infoInterval.Restart();
            }

            if (!PLCos.IsConnectedWithPLC)//连接断开
                return;

            if (Accessory.Velocity < 2)//没有生产
                return;

            //记录数据
            Db_WinderInfo db_WinderInfo = new Db_WinderInfo
            {
                Time = DateTime.Now,
                Velocity = Math.Round(Accessory.Velocity, 1),
                Traction1Current = Math.Round(Accessory.Traction1Current, 1),
                Traction2Current = Math.Round(Accessory.Traction2Current, 1),
                Winder0Current = Math.Round(Items[0].Current,1),
                Winder1Current = Math.Round(Items[1].Current, 1),
                Traction2TensionKg = Math.Round(Accessory.Traction2TensionKg,1),
                Winder0TensionKg = Math.Round(Items[0].TensionKg, 1),
                Winder1TensionKg = Math.Round(Items[1].TensionKg, 1),
                RotaryCurrent = Math.Round(Accessory.RotaryCurrent,1),
                RotaryFreq = Math.Round(Accessory.RotaryFreq,1),
                IsRotaryForw = Accessory.IsRotaryForw,
                IsRotaryBackw = Accessory.IsRotaryBackw
            };

            //写入数据库
            if (!stopwatch_1min.IsRunning)
            {
                //第1次立刻写入
                mHistoryDb.AddWinderInfo(db_WinderInfo);
                stopwatch_1min.Start();
            }
            else
            {
                //第2次，1分钟后才能保存
                dbInfo_1min.Add(db_WinderInfo);
            }

        }
    }
}
